@@ -11,6 +11,7 @@ import sys
 import tarfile
 import threading
 import time
+import boto3
 import requests
 import multiprocessing
 from base64 import b64encode
@@ -187,7 +188,7 @@ def query_to_pandas(device, query_result, sort=True):
         device_data_pd = device_data_pd.sort_index(axis=1)
     return device_data_pd
 
-def query(query, experiment_id, host, params={}):
+def query(query, experiment_id, host):
     """
     Query the MDML for an example of the data structure that your query will return. This is aimed at aiding in development of FuncX functions for use with the MDML.
 
@@ -206,7 +207,7 @@ def query(query, experiment_id, host, params={}):
         Data structure that will be passed to FuncX
     """
     import json
-    resp = requests.get(f"http://{host}:1880/query?query={json.dumps(query)}&parameters={json.dumps(params)}&experiment_id={experiment_id}")
+    resp = requests.get(f"http://{host}:1880/query?query={json.dumps(query)}&experiment_id={experiment_id}")
     return json.loads(resp.text)
 
 class experiment:
@@ -226,7 +227,7 @@ class experiment:
     
     """
 
-    def __init__(self, experiment_id, username, passwd, host):
+    def __init__(self, experiment_id, username, passwd, host, s3_access_key=None, s3_secret_key=None):
         
         self.experiment_id = experiment_id.upper()
         self.username = username
@@ -237,7 +238,6 @@ class experiment:
         self.msg_queue = None
         self.debugger = None
         self.debug_callback = None
-        
 
         # Creating connection to MQTT broker
         client = mqtt.Client()
@@ -257,6 +257,19 @@ class experiment:
         except:
             print("ERROR! Could not connect to the MDML's message broker. Verify you have the correct host. Contact jelias@anl.gov if the problem persists.")
 
+        if s3_access_key is not None:
+            # Creating boto3 (s3) client connection
+            session = boto3.session.Session()
+            try:
+                self.s3_client = session.client(
+                    service_name='s3',
+                    aws_access_key_id=s3_access_key,
+                    aws_secret_access_key=s3_secret_key,
+                    endpoint_url='https://s3.it.anl.gov:18082'
+                )
+            except:
+                print("ERROR! Connection to s3 failed.")
+
     def globus_login(self):
         """
         Perform a Globus login to acquire auth tokens for FuncX analyses. Must be done before experiment.add_config()
@@ -495,7 +508,7 @@ class experiment:
         # Send data via MQTT
         self.client.publish(topic, json.dumps(payload))
 
-    def use_dlhub(self, data, device_id, function_id):
+    def use_dlhub(self, data, device_id, function_id, funcx_callback=None):
         """
         Publish a message to run an analysis
 
@@ -520,6 +533,15 @@ class experiment:
             'parameters': {"data": data}
         }
 
+        # Add FuncX callback function
+        if funcx_callback is not None:
+            # Test format of the inputs
+            assert "endpoint_uuid" in funcx_callback.keys()
+            assert "function_uuid" in funcx_callback.keys()
+            assert "save_intermediate" in funcx_callback.keys()
+            # Add to payload
+            payload['funcx_callback'] = funcx_callback
+
         # Add auth if set
         if self.tokens:
             payload['globus_token'] = self.tokens['funcx_service']['access_token']
@@ -531,6 +553,55 @@ class experiment:
         # Send data via MQTT
         self.client.publish(topic, json.dumps(payload))
 
+    def publish_image_s3(self, device_id, img_filename, timestamp = 0, metadata = {}):
+        """
+        Publish an image to MDML
+
+        Parameters
+        ----------
+        device_id : str
+            Unique string identifying the device that created the data.
+            This must correspond with the experiment's configuration
+        img_filename : str
+            filename to store the file in the MDML. Can only contain letters, 
+            numbers, underscores and must end with a valid file extension. 
+            If left blank, filenames will the experiment ID followed by an
+            index (e.g. EXPID_1.JPG, EXPID_2.JPG...)
+        timestamp : int
+            Unix time in nanoseconds. Can be supplied by the unix_time()
+            function in this package
+        metadata : dict
+            Dictionary containing any metadata for the image. Data types of 
+            the dictionary values must not be changed. Keys cannot include
+            "time" or "filepath".
+        """
+
+        # Creating MQTT topic
+        topic = "MDML/" + self.experiment_id + "/DATA/" + device_id.upper()
+        # Base payload
+        payload = {
+            'filename': img_filename,
+            'data_type': 'image'
+        }
+        # Adding metadata if necessary
+        payload['metadata'] = metadata
+        # # Check for valid filename
+        # if img_filename != '':
+        #     if re.match(r"^[\w/-]+\.[A-Za-z0-9]+$", img_filename) == None:
+        #         print("Filename not valid. Can only contains letters, numbers, and underscores.")
+        #         return
+        #     else:
+        #         payload['filename'] = img_filename
+        payload['filename'] = img_filename
+        # Adding timing code
+        if timestamp == 0:
+            timestamp = unix_time()
+        payload['timestamp'] = timestamp
+        payload = json.dumps(payload)
+        # Upload to s3
+        self.s3_client.upload_file(img_filename, f'mdml-{self.experiment_id.lower()}', img_filename)
+        # Publish it
+        self.client.publish(topic, payload)
         
     def publish_image(self, device_id, img_byte_string, filename = '', timestamp = 0, metadata = {}):
         """
@@ -584,7 +655,7 @@ class experiment:
         # Publish it
         self.client.publish(topic, payload)
 
-    def query(self, query, host):
+    def query(self, query):
         """
         Query the MDML for an example of the data structure that your query will return. This is aimed at aiding in development of FuncX functions for use with the MDML.
 
@@ -603,7 +674,7 @@ class experiment:
             Data structure that will be passed to FuncX
         """
         import json
-        resp = requests.get(f"http://{host}:1880/query?query={json.dumps(query)}&experiment_id={self.experiment_id}")
+        resp = requests.get(f"http://{self.host}:1880/query?query={json.dumps(query)}&experiment_id={self.experiment_id}")
         return json.loads(resp.text)
 
     def _publish_image_benchmarks(self, device_id, img_byte_string, filename = '', timestamp = 0, size = 0):
@@ -654,13 +725,22 @@ class experiment:
         # Publish it
         self.client.publish(topic, payload)
                 
-    def reset(self):
+    def reset(self, hard_reset=False):
         """
         Publish a reset message on the MDML message broker to reset
         your current experiment.
+
+        Parameters
+        ----------
+        hard_reset : bool
+            True if the experiment should be ended regardless of analysis 
+            progress.            
         """
         topic = "MDML/" + self.experiment_id + "/RESET"
-        self.client.publish(topic, '{"reset": 1}')
+        if hard_reset:
+            self.client.publish(topic, '{"reset": 1, "hard_reset": 1}')
+        else:
+            self.client.publish(topic, '{"reset": 1}')
     
     def start_debugger(self):
         """
